@@ -2,26 +2,30 @@
 #
 #  build-initrd.sh
 #
-#  Written by Jari Ruusu, September 23 2006
+#  Written by Jari Ruusu, September 5 2021
 #
-#  Copyright 2001-2006 by Jari Ruusu.
+#  Copyright 2001-2021 by Jari Ruusu.
 #  Redistribution of this file is permitted under the GNU Public License.
 #
-#  Changes by Hauke Johannknecht <ash@ash.de> 11/2001
+#  Changes by Hauke Johannknecht 2001, 2009
 #     - added Pivot mode
 #     - added conffile-loading
 #     - added initrdonly-mode
 #     - added device 0 nodes
+#     - added init-md mode 
 #
 #
-#  Initrd can use two different methods to switch to encrypted root device:
-#  change_root (USEPIVOT=0) and pivot_root (USEPIVOT=1). change_root method
-#  is present in at least 2.2 and 2.4 kernels, and it works ok. pivot_root
-#  is present in 2.4 and later kernels, and offers much nicer wrong password
-#  case handling because initrd code can properly shutdown the kernel.
-#  Proper shutdown is important for software RAID devices and such.
-#  change_root and pivot_root require slightly different kernel and
-#  bootloader configuration.
+#  Initrd can use three different methods to switch to encrypted root
+#  device: change_root (USEPIVOT=0), pivot_root (USEPIVOT=1) and
+#  initramfs/switch_root (USEPIVOT=2). change_root method is available on at
+#  least 2.2, 2.4 and 2.6 kernels, and it works ok. pivot_root method is
+#  available on 2.4 and later kernels, and offers much nicer wrong
+#  passphrase case handling because initrd code can properly shutdown the
+#  kernel. initramfs/switch_root method is available on 2.6.13 and later
+#  kernels, and is similar to pivot_root. Proper shutdown is important for
+#  software RAID devices and such. change_root, pivot_root, and
+#  initramfs/switch_root require slightly different kernel and bootloader
+#  configurations.
 #
 #  kernel .config :  CONFIG_BLK_DEV_RAM=y
 #  (USEPIVOT=0)      CONFIG_BLK_DEV_RAM_SIZE=4096
@@ -35,6 +39,9 @@
 #                    CONFIG_BLK_DEV_INITRD=y
 #                    CONFIG_MINIX_FS=y
 #
+#  kernel .config :  CONFIG_BLK_DEV_INITRD=y
+#  (USEPIVOT=2)
+#
 #  /etc/lilo.conf :  initrd=/boot/initrd.gz
 #  (USEPIVOT=0)      root=/dev/ram1
 #                                 ^
@@ -42,18 +49,30 @@
 #  (USEPIVOT=1)      initrd=/boot/initrd.gz
 #                    root=/dev/ram0
 #                                 ^
+#  /etc/lilo.conf :  initrd=/boot/initrd.gz
+#  (USEPIVOT=2)
+#
 #  /boot/grub/menu.lst :  root (hd0,0)
-#  (USEPIVOT=0)           kernel /vmlinuz ro root=101
-#                         initrd /initrd.gz         ^
+#  (USEPIVOT=0)           kernel /vmlinuz root=101
+#                         initrd /initrd.gz      ^
 #
 #  /boot/grub/menu.lst :  root (hd0,0)
 #  (USEPIVOT=1)           kernel /vmlinuz root=100 init=/linuxrc rootfstype=minix
 #                         initrd /initrd.gz      ^
 #
+#  /boot/grub/menu.lst :  root (hd0,0)
+#  (USEPIVOT=2)           kernel /vmlinuz
+#                         initrd /initrd.gz
+#
 #  usage :  ./build-initrd.sh [configfile]
 #           lilo
 #           mkdir /boot/modules-`uname -r`
 #           cp -p /lib/modules/`uname -r`/block/loop.*o /boot/modules-`uname -r`/
+#  or                                     ^^^^^
+#           cp -p /lib/modules/`uname -r`/extra/loop.*o /boot/modules-`uname -r`/
+#                                         ^^^^^
+#  2.4 and older kernels always install to block/ directory
+#  2.6 kernels with loop-AES-v3.2a and later install to extra/ directory
 #
 
 ### All default-values can be altered via the configfile
@@ -63,10 +82,10 @@
 # must be configured with CONFIG_DEVFS_FS=y CONFIG_DEVFS_MOUNT=y
 USEDEVFS=0
 
-# 1 = use pivot_root, 0 = use old change_root
+# 0 = use old change_root, 1 = use pivot_root, 2 = use initramfs/switch_root
 # See above header for root= and append= lilo.conf definitions.
 # pivot_root is not available on 2.2 and older kernels.
-# Always enable pivot_root for 2.4 and later kernels.
+# initramfs/switch_root is not available on kernels older than 2.6.13
 USEPIVOT=1
 
 # Unencrypted /boot partition. If devfs is enabled (USEDEVFS=1), this must
@@ -156,11 +175,26 @@ TEMPLOOPINDEX=7
 # Example: LOOPMODPARAMS="max_loop=8 lo_prealloc=125,5,200"
 LOOPMODPARAMS=""
 
+# 1 = set keyboard to UTF-8 mode, 0 = don't set
+UTF8KEYBMODE=0
+
 # 1 = load national keyboard layout, 0 = don't load
 # You _must_ manually copy correct keyboard layout to /boot/default.kmap
 # which must be in uncompressed form. (can not be .gz file)
 LOADNATIONALKEYB=0
 
+# Try to auto-assemble linux software raid md devices. This is only
+# needed and used on USEPIVOT=2 (initramfs/switch_root) type build.
+# This gets automatically disabled if none of needed devices (BOOTDEV,
+# CRYPTROOT, or EXTERNALGPGDEV) is a /dev/md* device.
+# 1 = auto-assemble, 0 = no
+INITMD=1
+
+# Delay in seconds before /linuxrc attempts to auto-assemble
+# linux software raid md devices. This is only needed and used
+# on USEPIVOT=2 (initramfs/switch_root) type build.
+INITMDDELAY=1
+
 # Initial delay in seconds before /linuxrc attempts to mount /boot
 # partition. Slow devices (USB-sticks) may need some delay.
 INITIALDELAY=0
@@ -178,12 +212,17 @@ TOOLSPROMPT=0
 # If enabled, rootsetup program (+libs) _must_ be manually copied to /boot.
 USEROOTSETUP=0
 
-# 1 = use dietlibc to build linuxrc. This permits passing parameters to init.
-# 0 = use glibc to build linuxrc. This prevents passing parameters to init
+# 1 = use dietlibc to build /linuxrc. This permits passing parameters to init.
+# 0 = use glibc to build /linuxrc. This prevents passing parameters to init
 # and includes hacks that may be incompatible with some versions of glibc.
 # The dietlibc can be found at http://www.fefe.de/dietlibc/
 USEDIETLIBC=1
 
+# C compiler used to compile /linuxrc program.
+# 32bit x86 ubuntu-7.04 gcc-4.1.2 is known to miscompile /linuxrc. Affected
+# users should install gcc-3.3 package, and change this to GCC=gcc-3.3
+GCC=gcc
+
 # 1 = load extra module, 0 = don't load
 # If this is enabled, module must be manually copied to
 # /boot/modules-KERNELRELEASE/ directory under name like foomatic.o
@@ -207,6 +246,39 @@ EXTRAMODULELOAD5=0
 EXTRAMODULENAME5=""
 EXTRAMODULEPARAMS5=""
 
+# 1 = run extra command on encrypted root before starting init,
+# 2 = run + show command, 0 = don't run
+# If you set up loop devices, static loop and backing device nodes must
+# exist on /dev directory on read-only mounted encrypted root file system.
+# If needed, you can create them manually by booting to rescue
+# floppy/CD-ROM, mounting the file system, and using mknod program to create
+# static device nodes. Like this:
+# mknod -m 660 /mnt/dev/loop6 b 7 6
+# mknod -m 660 /mnt/dev/md3 b 9 3
+EXTRACOMMANDRUN1=0
+EXTRACOMMANDSTR1="/sbin/losetup -e AES128 -P /etc/cleartextkey-loop6.txt /dev/loop6 /dev/md3"
+# 1 = run extra command, 2 = run + show command, 0 = don't run
+EXTRACOMMANDRUN2=0
+EXTRACOMMANDSTR2=""
+# 1 = run extra command, 2 = run + show command, 0 = don't run
+EXTRACOMMANDRUN3=0
+EXTRACOMMANDSTR3=""
+# 1 = run extra command, 2 = run + show command, 0 = don't run
+EXTRACOMMANDRUN4=0
+EXTRACOMMANDSTR4=""
+# 1 = run extra command, 2 = run + show command, 0 = don't run
+EXTRACOMMANDRUN5=0
+EXTRACOMMANDSTR5=""
+# 1 = run extra command, 2 = run + show command, 0 = don't run
+EXTRACOMMANDRUN6=0
+EXTRACOMMANDSTR6=""
+# 1 = run extra command, 2 = run + show command, 0 = don't run
+EXTRACOMMANDRUN7=0
+EXTRACOMMANDSTR7=""
+# 1 = run extra command, 2 = run + show command, 0 = don't run
+EXTRACOMMANDRUN8=0
+EXTRACOMMANDSTR8=""
+
 ### End of options
 
 
@@ -244,6 +316,19 @@ if [ ${EXTERNALGPGFILES} == 1 ] ; then
     GPGMNTPATH=mnt
 fi
 
+INITMD_ENABLED=0
+if [ ${INITMD} == 1 ] ; then
+    for x in ${BOOTDEV} ${CRYPTROOT} ${GPGMOUNTDEV} ; do
+        case ${x} in
+            /dev/md[0-9]|/dev/md[0-9][0-9])
+                INITMD_ENABLED=1
+                INITMD_NAME=${x}
+                break
+                ;;
+        esac
+    done
+fi
+
 if [ ${USEDIETLIBC} == 1 ] ; then
     x=`which diet`
     if [ x${x} == x ] ; then
@@ -273,7 +358,17 @@ cat - <<EOF >tmp-c-$$.c
 #include <fcntl.h>
 #include <time.h>
 
-#if ${USEPIVOT}
+#if ${INITMD_ENABLED} && ${USEPIVOT} == 2
+#include <sys/ioctl.h>
+#include <linux/major.h>
+#include <linux/raid/md_u.h>
+#endif
+
+#ifndef MS_MOVE
+# define MS_MOVE 8192
+#endif
+
+#if ${USEPIVOT} == 1
 # if ${USEDIETLIBC}
    extern int pivot_root(const char *, const char *);
 # else
@@ -377,6 +472,37 @@ void runInit()
     doHalt();
 }
 
+#if ${USEPIVOT} == 2
+void removeInitramfsFiles()
+{
+    int x;
+    x =  unlink("${BOOTDEV}");
+    x |= unlink("${CRYPTROOT}");
+#if ${EXTERNALGPGFILES}
+    x |= unlink("${GPGMOUNTDEV}");
+#endif
+    x |= unlink("/dev/console");
+    x |= unlink("/dev/tty");
+    x |= unlink("/dev/tty1");
+    x |= unlink("/dev/null");
+    x |= unlink("/dev/zero");
+    x |= unlink("/dev/ram0");
+    x |= unlink("/dev/ram1");
+    x |= unlink("/dev/loop${ROOTLOOPINDEX}");
+    x |= rmdir("/dev");
+    x |= rmdir("/lib");
+    x |= unlink("/lib64");
+#if ${EXTERNALGPGFILES}
+    x |= rmdir("/mnt");
+#endif
+#if ${USEGPGKEY}
+    x |= unlink("/bin");
+#endif
+    x |= unlink("/init");
+    if(x) wrStr("removing initramfs files failed. (not fatal)\n");
+}
+#endif
+
 #if ${USEDIETLIBC}
 int main(int argc, char **argv)
 #else
@@ -384,8 +510,11 @@ void _start()
 #endif
 {
     int x = 0;
+    char buf[1000];
+#if ${USEMODULE} || ${EXTRAMODULELOAD1} || ${EXTRAMODULELOAD2} || ${EXTRAMODULELOAD3} || ${EXTRAMODULELOAD4} || ${EXTRAMODULELOAD5}
     struct utsname un;
-    char buf[1000], *modext;
+    char *modext;
+#endif
 
 #if ${USEDIETLIBC}
     argv_init = argv;
@@ -395,24 +524,52 @@ void _start()
     if(getpid() != 1) {
         /* pivot_root was configured, but kernel has */
         /* wandered off to change_root code path!    */
-        wrStr("ERROR: initrd config says USEPIVOT=1, but bootloader acts like USEPIVOT=0\n");
+        wrStr("ERROR: initrd config says USEPIVOT>0, but bootloader acts like USEPIVOT=0\n");
         _exit(0);
     }
 #else
     if(getpid() == 1) {
         /* change_root was configured, but kernel    */
         /* has wandered off to pivot_root code path! */
-        wrStr("ERROR: initrd config says USEPIVOT=0, but bootloader acts like USEPIVOT=1\n");
+        wrStr("ERROR: initrd config says USEPIVOT=0, but bootloader acts like USEPIVOT>0\n");
         runInit();
     }
 #endif
 
+#if ${INITMD_ENABLED} && ${USEPIVOT} == 2
+#if ${INITMDDELAY}
+    {
+        struct timespec req;
+        wrStr("Delaying md autoassemble for ${INITMDDELAY} seconds...\n");
+        req.tv_sec = ${INITMDDELAY};
+        req.tv_nsec = 0;
+        nanosleep(&req, 0);
+        wrStr("...delay complete, continuing\n");
+    }
+#endif
+    {
+        int y;
+        if((y = open("${INITMD_NAME}", O_RDONLY|O_DIRECT, 0)) == -1) {
+            wrStr("Unable to open ${INITMD_NAME} for triggering autoassemble\n");
+            goto fail4;
+        }
+        if(ioctl(y, RAID_AUTORUN, 0)) {
+            wrStr("Unable to trigger md autoassemble\n");
+            close(y);
+            goto fail4;
+        }
+        close(y);
+    }
+#endif
+
 #if ${INITIALDELAY}
     {
         struct timespec req;
+        wrStr("Delaying ${BOOTDEV} mount for ${INITIALDELAY} seconds...\n");
         req.tv_sec = ${INITIALDELAY};
         req.tv_nsec = 0;
         nanosleep(&req, 0);
+        wrStr("...delay complete, continuing\n");
     }
 #endif
 
@@ -427,18 +584,29 @@ void _start()
         goto fail4;
     }
 
+#if ${UTF8KEYBMODE}
+    buf[0] = 0;
+    strCat(buf, "/lib/kbd_mode -u");
+    exeWait(buf);
+#endif
 #if ${LOADNATIONALKEYB}
     buf[0] = 0;
-    strCat(buf, "/lib/loadkeys /lib/default.kmap");
+    strCat(buf, "/lib/loadkeys");
+#if ${UTF8KEYBMODE}
+    strCat(buf, " -u");
+#endif
+    strCat(buf, " /lib/default.kmap");
     exeWait(buf);
 #endif
 
+#if ${USEMODULE} || ${EXTRAMODULELOAD1} || ${EXTRAMODULELOAD2} || ${EXTRAMODULELOAD3} || ${EXTRAMODULELOAD4} || ${EXTRAMODULELOAD5}
     uname(&un);
     if((un.release[0] > '2') || (un.release[1] != '.') || (un.release[2] >= '6') || (un.release[3] != '.')) {
         modext = ".ko";
     } else {
         modext = ".o";
     }
+#endif
 
 #if ${USEMODULE}
     buf[0] = 0;
@@ -500,9 +668,11 @@ void _start()
 #if ${MOUNTDELAY}
     {
         struct timespec req;
+        wrStr("Delaying ${EXTERNALGPGDEV} mount for ${MOUNTDELAY} seconds...\n");
         req.tv_sec = ${MOUNTDELAY};
         req.tv_nsec = 0;
         nanosleep(&req, 0);
+        wrStr("...delay complete, continuing\n");
     }
 #endif
     if(mount("${EXTERNALGPGDEV}", "/mnt", "${EXTERNALGPGTYPE}", MS_MGC_VAL | MS_RDONLY, 0)) {
@@ -534,7 +704,7 @@ void _start()
 
 #if ${USEPIVOT}
     if(mount("/dev/loop${DEVFSSLASH1}${ROOTLOOPINDEX}", "/new-root", "${ROOTTYPE}", MS_MGC_VAL | MS_RDONLY, 0)) {
-        wrStr("Looks like you didn't say the magic word. Mounting /dev/loop${DEVFSSLASH1}${ROOTLOOPINDEX} failed\n");
+        wrStr("Mounting /dev/loop${DEVFSSLASH1}${ROOTLOOPINDEX} failed\n");
         buf[0] = 0;
         strCat(buf, "/lib/${LOSETUPPROG} -d /dev/loop${DEVFSSLASH1}${ROOTLOOPINDEX}");
         if(exeWait(buf)) goto fail3;
@@ -549,6 +719,22 @@ void _start()
         wrStr("chdir() to /new-root failed\n");
         goto fail1;
     }
+
+#if ${USEPIVOT} == 2
+    removeInitramfsFiles();
+    if(mount(".", "/", 0, MS_MGC_VAL | MS_MOVE, 0)) {
+        wrStr("Overmounting root failed\n");
+        fail1:
+        chdir("/");
+        umount("/new-root");
+        goto fail4;
+    }
+    /* initramfs/switch_root type setup wants chroot() immediately after mount() */
+    if(chroot(".")) {
+        wrStr("chroot() to new root failed\n");
+        goto fail1;
+    }
+#else
     if(pivot_root(".", "initrd")) {
         wrStr("pivot_root() to new root failed.\n Either 'initrd' directory is missing from your encrypted root partition\n or your kernel doesn't have pivot_root system call.\n");
         fail1:
@@ -556,6 +742,8 @@ void _start()
         umount("/new-root");
         goto fail4;
     }
+    /* pivot_root type setup wants chdir("/") immediately after pivot_root() */
+#endif
     chdir("/");
 
 #if ${USEDEVFS}
@@ -575,13 +763,69 @@ void _start()
     dup2(x, 2);
     close(x);
 
+#if ${USEPIVOT} == 1
+    /* pivot_root type setup wants chroot() after chdir("/") */
     if(chroot(".")) {
         wrStr("chroot() to new root failed\n");
         goto fail1;
     }
     wrStr("Pivoting to encrypted root completed successfully\n");
-    runInit();
 #else
+    wrStr("Switching to encrypted root completed successfully\n");
+#endif
+
+#if ${EXTRACOMMANDRUN1}
+#if ${EXTRACOMMANDRUN1} == 2
+    wrStr("Running command: ${EXTRACOMMANDSTR1}\n");
+#endif
+    buf[0] = 0; strCat(buf, "${EXTRACOMMANDSTR1}"); exeWait(buf);
+#endif
+#if ${EXTRACOMMANDRUN2}
+#if ${EXTRACOMMANDRUN2} == 2
+    wrStr("Running command: ${EXTRACOMMANDSTR2}\n");
+#endif
+    buf[0] = 0; strCat(buf, "${EXTRACOMMANDSTR2}"); exeWait(buf);
+#endif
+#if ${EXTRACOMMANDRUN3}
+#if ${EXTRACOMMANDRUN3} == 2
+    wrStr("Running command: ${EXTRACOMMANDSTR3}\n");
+#endif
+    buf[0] = 0; strCat(buf, "${EXTRACOMMANDSTR3}"); exeWait(buf);
+#endif
+#if ${EXTRACOMMANDRUN4}
+#if ${EXTRACOMMANDRUN4} == 2
+    wrStr("Running command: ${EXTRACOMMANDSTR4}\n");
+#endif
+    buf[0] = 0; strCat(buf, "${EXTRACOMMANDSTR4}"); exeWait(buf);
+#endif
+#if ${EXTRACOMMANDRUN5}
+#if ${EXTRACOMMANDRUN5} == 2
+    wrStr("Running command: ${EXTRACOMMANDSTR5}\n");
+#endif
+    buf[0] = 0; strCat(buf, "${EXTRACOMMANDSTR5}"); exeWait(buf);
+#endif
+#if ${EXTRACOMMANDRUN6}
+#if ${EXTRACOMMANDRUN6} == 2
+    wrStr("Running command: ${EXTRACOMMANDSTR6}\n");
+#endif
+    buf[0] = 0; strCat(buf, "${EXTRACOMMANDSTR6}"); exeWait(buf);
+#endif
+#if ${EXTRACOMMANDRUN7}
+#if ${EXTRACOMMANDRUN7} == 2
+    wrStr("Running command: ${EXTRACOMMANDSTR7}\n");
+#endif
+    buf[0] = 0; strCat(buf, "${EXTRACOMMANDSTR7}"); exeWait(buf);
+#endif
+#if ${EXTRACOMMANDRUN8}
+#if ${EXTRACOMMANDRUN8} == 2
+    wrStr("Running command: ${EXTRACOMMANDSTR8}\n");
+#endif
+    buf[0] = 0; strCat(buf, "${EXTRACOMMANDSTR8}"); exeWait(buf);
+#endif
+    runInit();
+
+#else
+    /* USEPIVOT=0 configured, write new device number to real-root-dev */
     if(mount("none", "/proc", "proc", MS_MGC_VAL, 0)) {
         wrStr("Mounting /proc failed\n");
         goto fail3;
@@ -600,7 +844,9 @@ void _start()
 #if ${EXTERNALGPGFILES}
     umount("/mnt");
 #endif
+#if ${USEMODULE} || ${EXTERNALGPGFILES} || ${EXTRAMODULELOAD1} || ${EXTRAMODULELOAD2} || ${EXTRAMODULELOAD3} || ${EXTRAMODULELOAD4} || ${EXTRAMODULELOAD5}
     fail5:
+#endif
     umount("/lib");
     fail4:
 #if ${USEPIVOT}
@@ -611,38 +857,45 @@ void _start()
 EOF
 
 if [ ${USEDIETLIBC} == 1 ] ; then
-    diet gcc -Wall -O2 -s -static -pipe tmp-c-$$.c -o tmp-c-$$
+    diet ${GCC} -Wall -O2 -s -static -pipe tmp-c-$$.c -o tmp-c-$$
 else
-    gcc -Wall -O2 -s -static -nostartfiles -pipe tmp-c-$$.c -o tmp-c-$$
+    ${GCC} -Wall -O2 -s -static -nostartfiles -pipe tmp-c-$$.c -o tmp-c-$$
 fi
 rm -f tmp-c-$$.[co]
 
-x=`cat tmp-c-$$ | wc -c`
-y=`expr ${x} + 1023`
-x=`expr ${y} / 1024`
-y=`expr ${x} + 11`
-if [ ${x} -gt 7 ] ; then
-    y=`expr ${y} + 1`
-fi
-if [ ${x} -gt 519 ] ; then
-    y=`expr ${y} + 3`
-fi
-if [ ${USEGPGKEY} == 1 ] ; then
-    y=`expr ${y} + 1`
-fi
-if [ ${EXTERNALGPGFILES} == 1 ] ; then
-    y=`expr ${y} + 1`
-fi
-
-dd if=/dev/zero of=tmp-i-$$ bs=1024 count=${y}
-/sbin/mkfs -t minix -i 32 tmp-i-$$
 mkdir tmp-d-$$
-mount -t minix tmp-i-$$ tmp-d-$$ -o loop=/dev/loop${DEVFSSLASH2}${TEMPLOOPINDEX}
+if [ ${USEPIVOT} != 2 ] ; then
+    x=`cat tmp-c-$$ | wc -c`
+    y=`expr ${x} + 1023`
+    x=`expr ${y} / 1024`
+    y=`expr ${x} + 11`
+    if [ ${x} -gt 7 ] ; then
+        y=`expr ${y} + 1`
+    fi
+    if [ ${x} -gt 519 ] ; then
+        y=`expr ${y} + 3`
+    fi
+    if [ ${USEGPGKEY} == 1 ] ; then
+        y=`expr ${y} + 1`
+    fi
+    if [ ${EXTERNALGPGFILES} == 1 ] ; then
+        y=`expr ${y} + 1`
+    fi
+
+    dd if=/dev/zero of=tmp-i-$$ bs=1024 count=${y}
+    /sbin/mkfs -t minix -i 32 tmp-i-$$ ${y}
+    mount -t minix tmp-i-$$ tmp-d-$$ -o loop=/dev/loop${DEVFSSLASH2}${TEMPLOOPINDEX}
+fi
 cd tmp-d-$$
 
 mkdir dev lib
 ln -s lib lib64
-mv ../tmp-c-$$ linuxrc
+
+if [ ${USEPIVOT} != 2 ] ; then
+    mv ../tmp-c-$$ linuxrc
+else
+    mv ../tmp-c-$$ init
+fi
 
 if [ ${EXTERNALGPGFILES} == 1 ] ; then
     mkdir mnt
@@ -652,7 +905,7 @@ if [ ${USEGPGKEY} == 1 ] ; then
     ln -s lib bin
 fi
 
-if [ ${USEPIVOT} == 1 ] ; then
+if [ ${USEPIVOT} != 0 ] ; then
     mkdir new-root
 else
     mkdir proc
@@ -680,25 +933,39 @@ if [ ${USEDEVFS} == 0 ] ; then
         mknodRefCount=`expr ${mknodRefCount} + 1`
     fi
 
-    maybeMakeDiskNode hda  3   0 ""
-    maybeMakeDiskNode hdb  3  64 ""
-    maybeMakeDiskNode hdc 22   0 ""
-    maybeMakeDiskNode hdd 22  64 ""
-    maybeMakeDiskNode hde 33   0 ""
-    maybeMakeDiskNode hdf 33  64 ""
-    maybeMakeDiskNode hdg 34   0 ""
-    maybeMakeDiskNode hdh 34  64 ""
-    maybeMakeDiskNode sda  8   0 ""
-    maybeMakeDiskNode sdb  8  16 ""
-    maybeMakeDiskNode sdc  8  32 ""
-    maybeMakeDiskNode sdd  8  48 ""
-    maybeMakeDiskNode sde  8  64 ""
-    maybeMakeDiskNode sdf  8  80 ""
-    maybeMakeDiskNode sdg  8  96 ""
-    maybeMakeDiskNode sdh  8 112 ""
-    maybeMakeDiskNode scd 11   0  0
-    maybeMakeDiskNode md   9   0  0
-    maybeMakeDiskNode fd   2   0  0
+    maybeMakeDiskNode hda       3   0 ""
+    maybeMakeDiskNode hdb       3  64 ""
+    maybeMakeDiskNode hdc      22   0 ""
+    maybeMakeDiskNode hdd      22  64 ""
+    maybeMakeDiskNode hde      33   0 ""
+    maybeMakeDiskNode hdf      33  64 ""
+    maybeMakeDiskNode hdg      34   0 ""
+    maybeMakeDiskNode hdh      34  64 ""
+    maybeMakeDiskNode sda       8   0 ""
+    maybeMakeDiskNode sdb       8  16 ""
+    maybeMakeDiskNode sdc       8  32 ""
+    maybeMakeDiskNode sdd       8  48 ""
+    maybeMakeDiskNode sde       8  64 ""
+    maybeMakeDiskNode sdf       8  80 ""
+    maybeMakeDiskNode sdg       8  96 ""
+    maybeMakeDiskNode sdh       8 112 ""
+    maybeMakeDiskNode uba     180   0 ""
+    maybeMakeDiskNode ubb     180   8 ""
+    maybeMakeDiskNode ubc     180  16 ""
+    maybeMakeDiskNode ubd     180  24 ""
+    maybeMakeDiskNode ube     180  32 ""
+    maybeMakeDiskNode ubf     180  40 ""
+    maybeMakeDiskNode ubg     180  48 ""
+    maybeMakeDiskNode ubh     180  56 ""
+    maybeMakeDiskNode scd      11   0  0
+    maybeMakeDiskNode sr       11   0  0
+    maybeMakeDiskNode md        9   0  0
+    maybeMakeDiskNode fd        2   0  0
+    maybeMakeDiskNode idac0d0p 72   0  0
+    maybeMakeDiskNode idac0d1p 72  16  0
+    maybeMakeDiskNode idac0d2p 72  32  0
+    maybeMakeDiskNode idac0d3p 72  48  0
+    maybeMakeDiskNode mmcblk0p 179  0  0
 
     if [ ${mknodCount} != ${mknodRefCount} ] ; then
         echo "*****************************************************************"
@@ -709,12 +976,19 @@ if [ ${USEDEVFS} == 0 ] ; then
         echo "***  Script aborted.                                          ***"
         echo "*****************************************************************"
         cd ..
-        umount tmp-d-$$
-        rmdir tmp-d-$$
-        rm tmp-i-$$
+        if [ ${USEPIVOT} != 2 ] ; then
+            umount tmp-d-$$
+            rmdir tmp-d-$$
+            rm tmp-i-$$
+        else
+            rm -rf tmp-d-$$
+        fi
         exit 1
     fi
 
+    # NOTE: If you add/remove/change these device names, then also edit
+    # removeInitramfsFiles() function so it can remove all device nodes
+
     mknod dev/console c 5 1
     mknod dev/tty c 5 0
     mknod dev/tty1 c 4 1
@@ -725,12 +999,18 @@ if [ ${USEDEVFS} == 0 ] ; then
     mknod dev/loop${ROOTLOOPINDEX} b 7 ${ROOTLOOPINDEX}
 fi
 
-cd ..
-df tmp-d-$$
-umount tmp-d-$$
-rmdir tmp-d-$$
-sync ; sync ; sync
-gzip -9 tmp-i-$$
+if [ ${USEPIVOT} != 2 ] ; then
+    cd ..
+    df tmp-d-$$
+    umount tmp-d-$$
+    rmdir tmp-d-$$
+    sync ; sync ; sync
+    gzip -9 tmp-i-$$
+else
+    find | cpio -o -H newc | gzip -9c >../tmp-i-$$.gz
+    cd ..
+    rm -rf tmp-d-$$
+fi
 mv tmp-i-$$.gz ${DESTINATIONROOT}${DESTINATIONPREFIX}/${INITRDGZNAME}
 ls -l ${DESTINATIONROOT}${DESTINATIONPREFIX}/${INITRDGZNAME}
 
@@ -750,6 +1030,9 @@ if [ ${USEMODULE}${EXTRAMODULELOAD1}${EXTRAMODULELOAD2}${EXTRAMODULELOAD3}${EXTR
         z="${z} /sbin/insmod.old"
     fi
 fi
+if [ ${UTF8KEYBMODE} == 1 ] ; then
+    z="${z} "`which kbd_mode`
+fi
 if [ ${LOADNATIONALKEYB} == 1 ] ; then
     z="${z} "`which loadkeys`
 fi
@@ -757,17 +1040,25 @@ if [ ${USEGPGKEY} == 1 ] ; then
     z="${z} "`which gpg`
 fi
 for x in ${z} ; do
-    echo Copying ${SOURCEROOT}${x} to ${DESTINATIONROOT}${DESTINATIONPREFIX}
-    cp -p ${SOURCEROOT}${x} ${DESTINATIONROOT}${DESTINATIONPREFIX}
-    y=`ldd ${SOURCEROOT}${x} | perl -ne 'if(/([^ ]*) \(0x/){print "$1\n"}'`
+    OUTFILE=`echo ${x} | sed 's%.*/%%'`
+    if [ -x ${SOURCEROOT}${x}.static ]; then
+      SOURCEFILE=${SOURCEROOT}${x}.static
+    else
+      SOURCEFILE=${SOURCEROOT}${x}
+    fi
+    echo Copying ${SOURCEFILE} to ${DESTINATIONROOT}${DESTINATIONPREFIX}/${OUTFILE}
+    cp -p ${SOURCEFILE} ${DESTINATIONROOT}${DESTINATIONPREFIX}/${OUTFILE} || echo "ERROR: cp command returned error status. Continuing anyway..."
+    y=`ldd ${DESTINATIONROOT}${DESTINATIONPREFIX}/${OUTFILE} | perl -ne 'if(/([^ ]*) \(0x/){print "$1\n"}'`
     for a in ${y} ; do
         echo Copying ${SOURCEROOT}${a} to ${DESTINATIONROOT}${DESTINATIONPREFIX}
-        cp -p ${SOURCEROOT}${a} ${DESTINATIONROOT}${DESTINATIONPREFIX}
+        cp -p ${SOURCEROOT}${a} ${DESTINATIONROOT}${DESTINATIONPREFIX} || echo "ERROR: cp command returned error status. Continuing anyway..."
     done
 done
 
-if [ ! -d ${DESTINATIONROOT}/initrd ] ; then
-    mkdir ${DESTINATIONROOT}/initrd
+if [ ${USEPIVOT} != 2 ] ; then
+    if [ ! -d ${DESTINATIONROOT}/initrd ] ; then
+        mkdir ${DESTINATIONROOT}/initrd
+    fi
 fi
 
 echo Done.
